//Justin Fiedler
//Project 1

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#define BUF_SIZE  	512
#define TRUE			1
#define FALSE		0

typedef int bool;
const bool DEBUG = 0;
enum {M_INTERACTIVE, M_BATCH, S_SEQUENTIAL, S_PARALLEL};

int mode;
int style;
char line[BUF_SIZE + 10];
char *cmd_strt;

typedef struct pcs_s *Processptr;

typedef struct pcs_s
{	int   qid;
	pid_t r_pid;
	int	argc;
	char *argv[BUF_SIZE];
} Process;

int pcsc;
Processptr pcsv[BUF_SIZE];
pid_t pidv[BUF_SIZE];

FILE *cin;

void init(void);
void prompt(void);
bool nextProcess(Processptr pcs);
int getProcesses(void);
Processptr newProcess(int qid);
void freeProcess(Processptr pcs);
void freeProcesses(void);
void setStyle(Processptr pcs);
void forkProcess(Processptr pcs);
void waitAllProcesses(void);
void printProcess(Processptr pcs);

int main(int argc, char *argv[])
{	init();
	if(argc == 1)		//INTERACTIVE
	{	mode = M_INTERACTIVE;
		cin = stdin;
	}
	else if(argc == 2)	//BATCH
	{	mode = M_BATCH;
		if((cin = fopen(argv[1], "r")) == NULL)
		{	fprintf(stderr, "error opening %s\n", argv[1]);
			exit(1);
		}	
	}
	else			//ERROR: ARG COUNT
	{	fprintf(stderr, "error: too many args\n"); fflush(stderr);
		exit(1);
	}

	while(TRUE)
	{	int i;
		prompt();
			if(DEBUG) {printf("%s\n", line); fflush(stdout);}

		getProcesses();
		for(i = 0; i < pcsc; i++)
		{	Processptr pcs = pcsv[i]; 
			if(strcmp(pcs->argv[0], "exit") == 0)
			{	if(DEBUG) printf("Exiting shell\n");
				if(style == S_PARALLEL) waitAllProcesses();
				freeProcesses();
				if(mode == M_BATCH)
				{	if(fclose(cin) != 0)
					{	fprintf(stderr, "error closing %s\n", argv[1]);
						exit(1);
					}
				}
				exit(0);
			}
			else if(strcmp(pcs->argv[0], "style") == 0)
			{	setStyle(pcs);
			}
			else
			{	forkProcess(pcs);
			}	
		}
		if(style == S_PARALLEL) waitAllProcesses();
		freeProcesses();
	}
	//!!! FILE ERRORS !!!
	if(mode == M_BATCH)	
	{	if(fclose(cin) != 0)
		{	fprintf(stderr, "error closing %s\n", argv[1]);
			exit(1);
		}
	}
}

/* init
 * initializes global variables
 */
void init(void)
{	style = S_SEQUENTIAL;
	cmd_strt = line;
	pcsc = 0;
	int i;
	for(i = 0; i < BUF_SIZE; i++)
	{	pidv[i] = 0;
	}
}

/* prompt
 * prompts user (if in interactive mode) and gets the current
 * input from the input stream (cin).
 */
void prompt(void)
{	//!!! INPUT ERRORS!!!
	int i;
	for(i = 0; i < (BUF_SIZE + 10); i++)
	{	line[i] = '\0';
	}
	
	if(mode == M_INTERACTIVE) printf("<ra-sh>: ");

	int c = 0;
	for(i = 0; i < BUF_SIZE && (char)c != '\n' && c != EOF; i++)
	{	if((c = fgetc(cin)) != EOF)
		{	line[i] = (char)c;
		}
		else
		{	if(feof(cin)) // EOF reached add exit command
			{	if(mode == M_BATCH) printf(line);
				strcat(line, ";exit");
				printf("\n"); fflush(stdout);
			}
			if(ferror(cin))
			{	fprintf(stderr, "%s", "input error\n"); fflush(stderr);
				exit(1);
			}
		}
	}
	
	if(c != EOF) line[i] = '\0';
	if(mode == M_BATCH) if(c != EOF) {printf(line); fflush(stdout);}
	
	if(i == BUF_SIZE)
	{	if(c != '\n' && c != EOF)
		{	fprintf(stderr, "overflow: max input %d characters\n", BUF_SIZE); fflush(stderr);
			while((c = fgetc(cin)) != '\n' && c != EOF);
			strcpy(line,";");
		}
	}
	strcat(line, ";");
}

/* nextProcess
 * parses input line and returns the next non-NULL argument
 * if there are no more arguments on the line it returns NULL
 */
bool nextProcess(Processptr pcs)
{	int i;
	bool end_of_line;
	char *token;
	char *cmd_end;
		
	do
	{	cmd_end = strchr(cmd_strt, ';');					//Find ';' if there is one
			//if(DEBUG) if(cmd_end != NULL){printf("cmd_end: %s\n", cmd_end); fflush(stdout);}
		end_of_line = (cmd_end == NULL) ? TRUE : FALSE;		//if there are no ';' then it is the last cmd
		if(cmd_end != NULL) *cmd_end = '\0';
			//if(DEBUG) {printf("cmd = %s|\n", cmd_strt); fflush(stdout);}

		//Convert command to args list !!! CHECK FOR MEMORY ERROR malloc return 0 !!!
		for(i = 0; (token = strtok(i == 0 ? cmd_strt : NULL, " \t\n")) != NULL; i++)
		{	pcs->argc++;
			pcs->argv[i] = (char *) malloc(strlen(token) * sizeof(char));
			strcpy(pcs->argv[i], token);
		}
		
		//if(DEBUG) printProcess(pcs);
		
		if(end_of_line)
		{	cmd_strt = line;		}
		else
		{	cmd_strt = cmd_end + 1;	}
			
	}while(pcs->argc == 0 && !end_of_line);
	
	if(pcs->argc != 0) pcsc++;
	return (pcs->argc != 0);
}

/* getProcesses
 * constructs a Process array of all the commands on the input line
 */
int getProcesses(void)
{	int i;
	
	// get new processes
	i = 0;
	do
	{	pcsv[i] = newProcess(i);
	}while(nextProcess(pcsv[i++]));

	//free last null process
	freeProcess(pcsv[pcsc]);
	pcsv[pcsc] = NULL;
	
	//print pcs list
	if(DEBUG)
	{	printf("Process List\n"); fflush(stdout);
		for(i = 0; i <= pcsc; i++)
		{	printProcess(pcsv[i]);
		}
	}
	return pcsc;
}

/* freeProcesses
 * frees the memory from all Processes in pcsv
 */
void freeProcesses()
{	int i;	
	for(i = 0; i < pcsc; i++)
	{	freeProcess(pcsv[i]);
		pcsv[i] = NULL;
	}
	pcsc = 0;
}

/* newProcess
 * allocates and returns a new Process and sets its pid to pid
 */
Processptr newProcess(int qid)
{	//!!! MEMORY ERRORS !!!
	int i;
	Processptr pcs = (Processptr) malloc(sizeof(Process));
	if(pcs == NULL)
	{	fprintf(stderr, "error: could not access memory"); fflush(stderr);
		exit(1);
	}
	pcs->qid = qid;
	pcs->r_pid = 0;
	pcs->argc = 0;
	for(i = 0; i < BUF_SIZE; i++)
	{	pcs->argv[i] = NULL;	
	}
	return pcs;
}

/* freeProcess
 * frees memory used by Process pcs
 */
void freeProcess(Processptr pcs)
{	if(DEBUG) printf("freeing process %d\n", pcs->qid); fflush(stdout);
	int i;
	for(i = 0; i < pcs->argc; i++)
	{	free(pcs->argv[i]);
		pcs->argv[i] = NULL;
	}
	free(pcs);
}

/* setStyle
 * sets the style to the given parameter or warns
 * the user on incorrect input
 */
void setStyle(Processptr pcs)
{	if(pcs->argv[1] == NULL)
	{	printf("style: too few args\n"); fflush(stdout);
	}
	else if(strcmp(pcs->argv[1], "sequential") == 0 || strcmp(pcs->argv[1], "s") == 0)
	{	style = S_SEQUENTIAL;
		waitAllProcesses();
	}
	else if(strcmp(pcs->argv[1], "parallel") == 0 || strcmp(pcs->argv[1], "p") == 0)
	{	style = S_PARALLEL;
	}
	else
	{	fprintf(stderr, "no such style: %s\n", pcs->argv[1]); fflush(stderr);
	}
	if(DEBUG) printf("style: %s", style == S_SEQUENTIAL ? "sequential\n" : "parallel\n");
}

/* forkProcess
 * forks a new process and execs it to the proper command
 */
void forkProcess(Processptr pcs)
{	pid_t pid;

	if((pid = fork()) < 0)
	{	fprintf(stderr, "error forking process: %s\n", pcs->argv[0]);
	}
	else if(pid == 0)		//Child
	{	if(execvp(pcs->argv[0], pcs->argv) == -1)
		{	fprintf(stderr, "%s: command not found\n", pcs->argv[0]);
		}
		exit(0);
	}
	else					//Parent
	{	pcs->r_pid = pid;
		if(style == S_SEQUENTIAL) waitAllProcesses();
		if(DEBUG) if(style == S_SEQUENTIAL) printf("Done waiting\n");
	}
}

void printProcess(Processptr pcs)
{	int i;
	if(pcs != NULL)
	{	printf("Process %d args %d\n", pcs->qid, pcs->argc); fflush(stdout);
		printf("pid: %d\n", pcs->r_pid); fflush(stdout);
		for(i = 0; i <= pcs->argc; i++)
		{	if(pcs->argv[i] == NULL)
			{	printf("args[%d] = NULL\n", i); fflush(stdout);
			}
			else
			{	printf("args[%d] = %s|\n", i, pcs->argv[i]); fflush(stdout);
			}
		}
	}
	else
	{	printf("Process NULL\n"); fflush(stdout);
	}
}

/* waitAllProcesses
 * waits for all processes to finish before unblocking the parent
 */
void waitAllProcesses(void)
{	pid_t ret_val;
	while((ret_val = wait(NULL)) > 0) continue;
	if((ret_val == -1) && (errno != 10))
	{	fprintf(stderr, "error: wait error"); fflush(stderr);
	} 
}

